
"""
Generic module broken out from guiplugins. Contains utility code that can be
called from anywhere in the gtkgui package
"""
import os
import operator
import locale
from texttestlib import plugins
from copy import copy
from functools import reduce

# Gtk.accelerator_valid appears utterly broken on Windows


def windowsAcceleratorValid(key, mod):
    name = Gtk.accelerator_name(key, mod)
    return len(name) > 0 and name != "VoidSymbol"


try:
    from gi.repository import Gtk
    if os.name == "nt":
        Gtk.accelerator_valid = windowsAcceleratorValid
except ImportError:
    pass  # We might want to document the config entries, silly to fail on lack of GTK...


def createApplicationEvent(name, category, **kw):
    try:
        from storytext import applicationEvent
        # Everything that comes from here is to do with editing files in external programs
        applicationEvent(name, category, **kw)
    except ImportError:
        pass


guiConfig = None


class Utf8Converter:
    langVars = ('LC_ALL', 'LC_CTYPE', 'LANG', 'LANGUAGE')

    def __init__(self):
        self.encodings = self.getEncodings()

    def getEncodings(self):
        encodings = ['utf-8']
        localeEncoding = locale.getdefaultlocale()[1]
        if localeEncoding and not localeEncoding in encodings:
            encodings.insert(0, localeEncoding)
        return encodings

    def convert(self, text, extraEncodingLookup=None):
        unicodeInfo = self.decodeText(text, extraEncodingLookup)
        return unicodeInfo.encode('utf-8', 'replace')

    def getExtraEncodings(self, extraEncodingLookup):
        # Calling undocumented method to save copying it, best not to crash if it has been renamed.
        # Don't do this on Windows, the variables concerned don't apply there anyway
        if extraEncodingLookup and os.name != "nt" and hasattr(locale, "_parse_localename"):
            # Copied from locale.getdefaultlocale, would be nice if we could actually call this code
            for variable in self.langVars:
                localename = extraEncodingLookup(variable)
                if localename:
                    if variable == 'LANGUAGE':
                        localename = localename.split(':')[0]
                    extraEncoding = locale._parse_localename(localename)[1]
                    if extraEncoding and extraEncoding not in self.encodings and extraEncoding.lower() not in self.encodings:
                        return [extraEncoding]
        return []

    def decodeText(self, text, extraEncodingLookup):
        encodingsToTry = self.encodings + self.getExtraEncodings(extraEncodingLookup)
        for encoding in encodingsToTry:
            try:
                return str(text, encoding, errors="strict")
            except Exception:
                pass

        return str(text, encodingsToTry[0], errors="replace")


def getImageDir():
    retro = guiConfig.getValue("retro_icons")
    currImageDir = plugins.installationDir("images")
    retroDir = os.path.join(currImageDir, "retro")
    return (retroDir, True) if retro != 0 else (currImageDir, False)


class RefreshTips:
    def __init__(self, name, refreshCell, refreshColumn, refreshIndex):
        self.name = name
        self.refreshIndex = refreshIndex
        self.refreshColumn = refreshColumn
        self.refreshCell = refreshCell

    def hasRefreshIcon(self, view, path):  # pragma: no cover - StoryText cannot test tooltips (future?)
        model = view.get_model()
        if isinstance(model, Gtk.TreeModelFilter):
            childPath = model.convert_path_to_child_path(path)
            return model.get_model()[childPath][self.refreshIndex]
        else:
            return model[path][self.refreshIndex]

    def getTooltip(self, view, widget_x, widget_y, dummy, tooltip):  # pragma: no cover - StoryText cannot test tooltips (future?)
        x, y = view.convert_widget_to_tree_coords(widget_x, widget_y)
        pathInfo = view.get_path_at_pos(x, y)
        if pathInfo is None:
            return False

        path, column, cell_x, _ = pathInfo
        if column is not self.refreshColumn or not self.hasRefreshIcon(view, path):
            return False

        cell_pos = column.cell_get_position(self.refreshCell)[0]
        if cell_x > cell_pos:
            tooltip.set_text("Indicates that this " + self.name + "'s approved result has changed since the status was calculated. " +
                             "It's therefore recommended to recompute the status.")
            return True
        else:
            return False


def addRefreshTips(view, *args):
    view.set_property("has-tooltip", True)
    refreshTips = RefreshTips(*args)
    view.connect("query-tooltip", refreshTips.getTooltip)


class GUIConfig:
    def __init__(self, dynamic, allApps, defaultColours, defaultAccelerators, includePersonal=True):
        self.apps = copy(allApps)
        self.dynamic = dynamic
        self.configDir = plugins.MultiEntryDictionary()
        self.configDocs = {}
        self.setConfigDefaults(defaultColours, defaultAccelerators)
        if includePersonal:
            self.configDir.readValues(self.getAllPersonalConfigFiles(), insert=0, errorOnUnknown=0)
        self.shownCategories = list(map(self.getConfigName, self.configDir.get("show_test_category")))
        self.hiddenCategories = list(map(self.getConfigName, self.configDir.get("hide_test_category")))
        self.colourDict = self.makeColourDictionary()

    def getAllPersonalConfigFiles(self):
        allPersonalFiles = []
        # Always include app-independent version
        appIndep = os.path.join(plugins.getPersonalConfigDir(), "config")
        if os.path.isfile(appIndep):
            allPersonalFiles.append(appIndep)
        for app in self.apps:
            for fileName in app.getPersonalConfigFiles():
                if not fileName in allPersonalFiles:
                    allPersonalFiles.append(fileName)
        return allPersonalFiles

    def addSuites(self, suites):
        fullNames = [app.fullName() for app in self.apps]
        for suite in suites:
            if suite.app.fullName() not in fullNames:
                self.apps.append(suite.app)

    def makeColourDictionary(self):
        d = {}
        if self.configDir.get("test_colours") is None:
            return d
        for key, value in list(self.configDir.get("test_colours").items()):
            d[self.getConfigName(key)] = value
        return d

    def setConfigDefaults(self, colourDict, accelerators):
        self.setConfigDefault("static_collapse_suites", 100,
                              "Starting at this level the static GUI will show the suites collapsed")
        self.setConfigDefault("test_colours", colourDict, "Colours to use for each test state")
        self.setConfigDefault("file_colours", copy(colourDict), "Colours to use for each file state")
        self.setConfigDefault("window_size", self.getWindowSizeSettings(),
                              "To set the initial size of the dynamic/static GUI.")
        self.setConfigDefault("hide_gui_element", self.getDefaultHideWidgets(), "List of widgets to hide by default")
        self.setConfigDefault("hide_test_category", [
                              "cancelled"], "Categories of tests which should not appear in the dynamic GUI test view")
        self.setConfigDefault("show_test_category", ["failed"],
                              "Categories of tests which should appear in the dynamic GUI test view")
        self.setConfigDefault("query_kill_processes", {"default": [], "static": [
                              "Dynamic GUI"]}, "Ask about whether to kill these processes when exiting texttest.")
        self.setConfigDefault("gui_accelerators", accelerators, "Custom action accelerators.")
        self.setConfigDefault("gui_entry_completion_matching", 1,
                              "Which matching type to use for entry completion. 0 means turn entry completions off, 1 means match the start of possible completions, 2 means match any part of possible completions")
        self.setConfigDefault("gui_entry_completion_inline", 0,
                              "Automatically inline common completion prefix in entry.")
        self.setConfigDefault("gui_entry_completions", {"default": []},
                              "Add these completions to the entry completion lists initially")
        self.setConfigDefault("sort_test_suites_recursively", 1, "Sort subsuites when sorting test suites")
        self.setConfigDefault("retro_icons", 0, "Use the old TextTest icons in the dynamic and static GUIs")

    def setConfigDefault(self, key, value, docString):
        self.configDir[key] = value
        self.configDocs[key] = docString

    def _simpleValue(self, app, entryName):
        return app.getConfigValue(entryName)

    def _compositeValue(self, app, *args, **kwargs):
        return app.getCompositeConfigValue(*args, **kwargs)

    def _getFromApps(self, method, *args, **kwargs):
        callables = [plugins.Callable(method, app, *args) for app in self.apps]
        aggregator = plugins.ResponseAggregator(callables)
        try:
            return aggregator(**kwargs)
        except plugins.AggregationError as e:
            app = self.apps[e.index]
            plugins.printWarning("GUI configuration '" + "::".join(args) +
                                 "' differs between applications, ignoring that from " + repr(app) + "\n" +
                                 "Value was " + repr(e.value2) + ", change from " + repr(e.value1), stdout=True)
            return e.value1

    def getModeName(self):
        if self.dynamic:
            return "dynamic"
        else:
            return "static"

    def getConfigName(self, name, modeDependent=False):
        formattedName = name.lower().replace(" ", "_").replace(":", "_")
        if modeDependent:
            if len(name) > 0:
                return self.getModeName() + "_" + formattedName
            else:
                return self.getModeName()
        else:
            return formattedName

    def getValue(self, entryName, modeDependent=False):
        nameToUse = self.getConfigName(entryName, modeDependent)
        guiValue = self.configDir.get(nameToUse)
        if guiValue is not None:
            return guiValue
        else:
            return self._getFromApps(self._simpleValue, nameToUse)

    def getCompositeValue(self, sectionName, entryName, modeDependent=False, defaultKey="default"):
        nameToUse = self.getConfigName(entryName, modeDependent)
        value = self.configDir.getComposite(sectionName, nameToUse, defaultKey=defaultKey)
        if value is None:
            value = self._getFromApps(self._compositeValue, sectionName, nameToUse, defaultKey=defaultKey)
        if modeDependent and value is None:
            return self.getCompositeValue(sectionName, entryName)
        else:
            return value

    def getWindowOption(self, name):
        return self.getCompositeValue("window_size", name, modeDependent=True)

    def getWindowDimension(self, dimensionName, diag):
        pixelDimension = self.getWindowOption(dimensionName + "_pixels")
        if pixelDimension != "<not set>":
            diag.info("Setting window " + dimensionName + " to " + pixelDimension + " pixels.")
            return int(pixelDimension)
        else:
            # needs review MB 2018-12-04
            from gi.repository import Gdk
            fullSize = getattr(Gdk.Screen.get_default(), "get_" + dimensionName)()
            proportion = float(self.getWindowOption(dimensionName + "_screen"))
            diag.info("Setting window " + dimensionName + " to " + repr(int(100.0 * proportion)) + "% of screen.")
            return int(fullSize * proportion)

    def showCategoryByDefault(self, category, parentShown=False):
        if self.dynamic:
            nameToUse = self.getConfigName(category)
            if nameToUse in self.hiddenCategories:
                return False
            elif nameToUse in self.shownCategories:
                return True
            else:
                return parentShown
        else:
            return False

    def getTestColour(self, category, fallback=None):
        if self.dynamic:
            nameToUse = self.getConfigName(category)
            if nameToUse in self.colourDict:
                return self.colourDict[nameToUse]
            elif fallback:
                return fallback
            else:
                return self.colourDict.get("failure")
        else:
            if category.startswith("clipboard"):
                return self.colourDict.get(category)
            else:
                return self.colourDict.get("static")

    @staticmethod
    def getWindowSizeSettings():
        d = {}
        # Ordering is important - type deduced from first entry
        # Most common that people vary height_screen and width_screen per mode, so make sure type ends up as float otherwise these will be rejected
        d["height_screen"] = float(5.0) / 6
        d["width_screen"] = 0.6
        d["horizontal_separator_position"] = 0.46
        d["vertical_separator_position"] = 0.5
        d["height_pixels"] = "<not set>"
        d["width_pixels"] = "<not set>"
        d["maximize"] = 0
        return d

    @staticmethod
    def getDefaultHideWidgets():
        d = {}
        d["status_bar"] = 0
        d["toolbar"] = 0
        d["shortcut_bar"] = 0
        return d

    @staticmethod
    def getDefaultColours():
        d = {}
        d["default"] = "salmon"
        d["success"] = "DarkSeaGreen2"
        d["failure"] = "salmon"
        d["running"] = "LightGoldenrod1"
        d["initial_filter"] = "LightGoldenrod1"
        d["final_filter"] = "LightGoldenrod1"
        d["not_started"] = "white"
        d["pending"] = "grey80"
        d["static"] = "grey90"
        d["clipboard_cut"] = "red"
        d["clipboard_copy"] = "grey60"
        d["bug"] = "orange"
        d["marked"] = "lightblue"
        return d


# base class for all "GUI" classes which manage parts of the display
class SubGUI(plugins.Observable):
    def __init__(self):
        plugins.Observable.__init__(self)
        self.widget = None

    def createView(self):  # pragma: no cover - implemented in all subclasses
        pass

    def shouldShow(self):
        return True  # should this be shown/created at all this run

    def shouldShowCurrent(self, *args):
        return True  # should this be shown or hidden in the current context?

    def getTabTitle(self):
        return ""

    def forceVisible(self, *args):
        return False

    def addScrollBars(self, view, hpolicy):
        window = Gtk.ScrolledWindow()
        window.set_policy(hpolicy, Gtk.PolicyType.AUTOMATIC)
        self.addToScrolledWindow(window, view)
        window.show()
        return window

    def addToScrolledWindow(self, window, widget):
        if isinstance(widget, Gtk.VBox):
            window.add_with_viewport(widget)
        else:
            window.add(widget)

    def applicationEvent(self, name, **kw):
        createApplicationEvent(name, "files", **kw)

# base class for managing containers


class ContainerGUI(SubGUI):
    def __init__(self, subguis):
        SubGUI.__init__(self)
        self.subguis = subguis

    def forceVisible(self, rowCount):
        return reduce(operator.or_, (subgui.forceVisible(rowCount) for subgui in self.subguis))

    def shouldShow(self):
        return reduce(operator.or_, (subgui.shouldShow() for subgui in self.subguis))

    def shouldShowCurrent(self, *args):
        return reduce(operator.and_, (subgui.shouldShowCurrent(*args) for subgui in self.subguis))

    def getTabTitle(self):
        return self.subguis[0].getTabTitle()



#!/usr/bin/env python3

# Interface to GitHub using the JSON API.

import os
import sys
import ssl
import urllib.request
import json


def findBugInfo(bugId, location, *args):
    if location and location[-1] != '/':
        location += '/'
    request = "%sissues/%s" % (location, bugId)
    try:
        if request.startswith("https") and getattr(sys, 'frozen', False):
            certs = os.path.join(os.path.dirname(sys.executable), "etc", "cacert.pem")
            reply = urllib.request.urlopen(request, context=ssl.create_default_context(cafile=certs))
        else:
            reply = urllib.request.urlopen(request)
        content =  reply.read().decode(reply.headers.get_content_charset())
        info = json.loads(content)
    except Exception as e:
        message = ("Failed to open URL '" + request + "': " + str(e) +
                   ".\n\nPlease make sure that bug " + bugId + " exists\n" +
                   "and that the configuration entry 'bug_system_location' " +
                   "points to the correct GitHub repository.\nThe current value is '" + location +
                   "', it often looks like: 'https://api.github.com/repos/<user>/<repo>/'.")
        return "NONEXISTENT", message, False, bugId
    if len(info) <= 1:
        message = "Could not parse reply from GitHub, maybe incompatible interface."
        return "BAD SCRIPT", message, False, bugId
    bugText = "******************************************************\n" + \
        "Ticket #%s (%s)\n" % (bugId, info['state']) + \
        "%s\n%sticket/%s\n" % (info['title'], location, bugId) + \
        "Reported By: %s Owned by: %s\n" % (info['user']['login'], info['assignee']) + \
        "Updated: %s Milestone: %s\n" % (info['updated_at'], info['milestone']) + \
        "Description:\n" + info['body'] + "\n" + \
        "******************************************************"
    return info['state'], bugText, info['state'] == "closed", bugId


if __name__ == "__main__":  # pragma: no cover - test code
    import sys
    for item in findBugInfo(sys.argv[1], sys.argv[2]):
        print(item)




"""
Utilities for both master and slave code
"""

import os
import socket
from texttestlib import plugins
from locale import getpreferredencoding

noReusePostfix = ".NO_REUSE"
rerunPostfix = ".RERUN_TEST"
sendFilePostfix = ".SEND_FILES"
getFilePostfix = ".GET_FILES"


def getIPAddress(apps):
    if useLocalQueueSystem(apps):
        return "127.0.0.1"  # always works if everything is local

    # Seems to be no good portable way to get the IP address in a portable way
    # See e.g. http://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
    # These two methods seem to be the only vaguely portable ones
    try:
        # Doesn't always work, sometimes not available
        return socket.gethostbyname(socket.gethostname())
    except socket.error:
        # Relies on being online, but seems there is no other way...
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 0))  # Google's DNS server. Should always be there :)
        return s.getsockname()[0]


def queueSystemName(app):
    return app.getConfigValue("queue_system_module")


def useLocalQueueSystem(apps):
    return all((queueSystemName(app) == "local" for app in apps))


def socketSerialise(test):
    return test.app.name + test.app.versionSuffix() + ":" + test.getRelPath()


def socketParse(testString):
    # Test name might contain ":"
    return testString.strip().split(":", 1)


def makeIdentifierLine(identifier, sendFiles=False, getFiles=False, noReuse=False, rerun=False):
    if sendFiles:
        identifier += sendFilePostfix
    if getFiles:
        identifier += getFilePostfix
    if noReuse:
        identifier += noReusePostfix
    if rerun:
        identifier += rerunPostfix
    return identifier


def parseIdentifier(line):
    rerun = line.endswith(rerunPostfix)
    if rerun:
        line = line.replace(rerunPostfix, "")

    tryReuse = not line.endswith(noReusePostfix)
    if not tryReuse:
        line = line.replace(noReusePostfix, "")

    sendFiles = line.endswith(sendFilePostfix)
    if sendFiles:
        line = line.replace(sendFilePostfix, "")

    getFiles = line.endswith(getFilePostfix)
    if getFiles:
        line = line.replace(getFilePostfix, "")

    return line, sendFiles, getFiles, tryReuse, rerun


dirText = "DIRECTORY_CONTENTS"
fileText = "FILE_CONTENTS"
endPrefix = "END_"


def directorySerialise(dirName, ignoreLinks=False):
    text = ""
    for root, _, files in os.walk(dirName):
        for fn in sorted(files):
            path = os.path.join(root, fn)
            if not os.path.islink(path):
                relpath = plugins.relpath(path, dirName)
                text += fileText + " " + relpath + "\n"
                with open(path) as f:
                    text += f.read()
                if not text.endswith("\n"):
                    text += "\n"
                text += endPrefix + fileText + "\n"
    text += endPrefix + dirText
    return text


def directoryUnserialise(rootDir, f):
    currFile = None
    for line in f:
        lineStr = str(line, getpreferredencoding())
        if currFile is not None:
            if lineStr.startswith(endPrefix + fileText):
                currFile.close()
                currFile = None
            else:
                currFile.write(lineStr)
        else:
            if lineStr.startswith(fileText):
                fn = lineStr.strip().split()[-1]
                path = os.path.join(rootDir, fn)
                plugins.ensureDirExistsForFile(path)
                currFile = open(path, "w")
            elif lineStr.startswith(endPrefix + dirText):
                break




# Support for our custom widgets for StoryText, which needs to know how to handle them

from gi.repository import Gtk
from storytext.gtktoolkit.simulator.treeviewevents import TreeColumnHelper
from storytext.gtktoolkit.simulator.baseevents import RightClickEvent


class TreeColumnRightClickEvent(RightClickEvent):
    def __init__(self, name, widget, argumentParseData):
        column = TreeColumnHelper.findColumn(widget, argumentParseData)
        RightClickEvent.__init__(self, name, column.get_button())

    @classmethod
    def canHandleEvent(cls, widget, signalName, argumentParseData):
        return argumentParseData and RightClickEvent.canHandleEvent(widget, signalName)

    @classmethod
    def getAssociatedSignatures(cls, widget):
        signatures = []
        for column in widget.get_columns():
            if hasattr(column, "get_button"):
                signatures.append(cls.signalName + "." + TreeColumnHelper.getColumnName(column))
        return signatures


customEventTypes = [(Gtk.TreeView, [TreeColumnRightClickEvent])]




import os
import shutil
import subprocess
import sys
from texttestlib import plugins
from locale import getpreferredencoding


class SetUpCaptureMockHandlers(plugins.Action):
    def __init__(self, recordSetting):
        self.recordSetting = recordSetting
        libexecDir = plugins.installationDir("libexec")
        self.siteCustomizeFile = os.path.join(libexecDir, "sitecustomize.py")

    def __call__(self, test):
        pythonCustomizeFiles = test.getAllPathNames("testcustomize.py")
        pythonCoverage = test.hasEnvironment("COVERAGE_PROCESS_START")
        captureMock = test.app.usesCaptureMock() and self.recordSetting != 3
        if captureMock or pythonCoverage or pythonCustomizeFiles:
            rcFiles = test.getAllPathNames("capturemockrc")
            if rcFiles or pythonCoverage or pythonCustomizeFiles:
                self.setUpIntercepts(test, rcFiles, pythonCoverage, pythonCustomizeFiles)
        elif self.recordSetting == 3:
            self.disableCaptureMockComparison(test.app)

    def disableCaptureMockComparison(self, app):
        for name in ["traffic", "externalmocks", "pythonmocks"]:
            app.removeConfigEntry("regenerate", name, "definition_file_stems")
            app.addConfigEntry("builtin", name, "definition_file_stems")

    def setUpIntercepts(self, test, rcFiles, pythonCoverage, pythonCustomizeFiles):
        interceptDir = test.makeTmpFileName("traffic_intercepts", forComparison=0)
        captureMockActive = False
        if rcFiles:
            captureMockActive = self.setUpCaptureMock(test, interceptDir, rcFiles)

        if pythonCustomizeFiles:
            self.intercept(pythonCustomizeFiles[-1], interceptDir)  # most specific

        useSiteCustomize = captureMockActive or pythonCoverage or pythonCustomizeFiles
        if useSiteCustomize:
            self.intercept(self.siteCustomizeFile, interceptDir)
            for var in ["PYTHONPATH", "JYTHONPATH"]:
                test.setEnvironment(var, interceptDir + os.pathsep + test.getEnvironment(var, ""))

    def setUpCaptureMock(self, test, interceptDir, rcFiles):
        extReplayFile = test.getFileName("traffic")
        if extReplayFile:
            # "Legacy" setup to avoid the need to rename hundreds of files
            extRecordFile = test.makeTmpFileName("traffic")
            pyReplayFile = extReplayFile
            pyRecordFile = extRecordFile
        else:
            extReplayFile = test.getFileName("externalmocks")
            extRecordFile = test.makeTmpFileName("externalmocks")
            pyReplayFile = test.getFileName("pythonmocks")
            pyRecordFile = test.makeTmpFileName("pythonmocks")
        recordEditDir = test.makeTmpFileName("file_edits", forComparison=0)
        replayEditDir = test.getFileName("file_edits") if extReplayFile else None
        sutDirectory = test.getDirectory(temporary=1, local=1)
        capturemock_path = test.getConfigValue("capturemock_path")
        if capturemock_path:
            sys.path.insert(0, capturemock_path)
        try:
            from capturemock import setUpServer, setUpPython
            externalActive = setUpServer(self.recordSetting, extRecordFile, extReplayFile,
                                         recordEditDir=recordEditDir, replayEditDir=replayEditDir,
                                         rcFiles=rcFiles, interceptDir=interceptDir,
                                         sutDirectory=sutDirectory, environment=test.environment)
            pythonActive = setUpPython(self.recordSetting, pyRecordFile, pyReplayFile, rcFiles=rcFiles,
                                       environment=test.environment)
            return externalActive or pythonActive
        except ImportError as e:
            raise plugins.TextTestError(
                "Test requires CaptureMock to be installed, but the capturemock module could not be found\nError was: " + str(e) + "\nSearched in " + repr(sys.path))

    def intercept(self, moduleFile, interceptDir):
        interceptName = os.path.join(interceptDir, os.path.basename(moduleFile))
        plugins.ensureDirExistsForFile(interceptName)
        self.copyOrLink(moduleFile, interceptName)

    def copyOrLink(self, src, dst):
        if os.name == "posix":
            os.symlink(src, dst)
        else:
            shutil.copy(src, dst)


class TerminateCaptureMockHandlers(plugins.Action):
    def __call__(self, test):
        try:
            from capturemock import terminate
            terminate()
        except ImportError:
            pass


class ModifyTraffic(plugins.ScriptWithArgs):
    scriptDoc = "Apply a script to all the client server data"

    def __init__(self, args):
        argDict = self.parseArguments(args, ["script", "types", "file"])
        self.script = argDict.get("script")
        self.fileName = argDict.get("file", "traffic")
        self.trafficTypes = plugins.commasplit(argDict.get("types", "CLI,SRV"))

    def __repr__(self):
        return "Updating CaptureMock files in"

    def __call__(self, test):
        fileName = test.getFileName(self.fileName)
        if not fileName:
            return

        self.describe(test)
        try:
            newTexts = [self.getModified(t, test.getDirectory()) for t in self.readIntoList(fileName)]
        except plugins.TextTestError as e:
            print(str(e).strip())
            return

        newFileName = fileName + "tmpedit"
        newFile = open(newFileName, "w")
        for text in newTexts:
            self.write(newFile, text)
        newFile.close()
        shutil.move(newFileName, fileName)

    def readIntoList(self, fileName):
        # Copied from CaptureMock ReplayInfo, easier than trying to reuse it
        trafficList = []
        currTraffic = ""
        for line in open(fileName, "rU"):
            if line.startswith("<-") or line.startswith("->"):
                if currTraffic:
                    trafficList.append(currTraffic)
                currTraffic = ""
            currTraffic += line
        if currTraffic:
            trafficList.append(currTraffic)
        return trafficList

    def getModified(self, fullLine, dir):
        trafficType = fullLine[2:5]
        args = [self.script, fullLine[6:]]
        if os.name == "nt":
            interpreter = plugins.getInterpreter(self.script)
            if interpreter:
                args = [interpreter] + args
        if trafficType in self.trafficTypes:
            proc = subprocess.Popen(args, cwd=dir, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = proc.communicate()
            if len(stderr) > 0:
                raise plugins.TextTestError("Couldn't modify traffic :\n " + str(stderr, getpreferredencoding()))
            else:
                return fullLine[:6] + str(stdout, getpreferredencoding())
        else:
            return fullLine

    def write(self, newFile, desc):
        if not desc.endswith("\n"):
            desc += "\n"
        newFile.write(desc)

    def setUpSuite(self, suite):
        self.describe(suite)


class ConvertToCaptureMock(plugins.Action):
    def convert(self, confObj, newFile):
        from configparser import ConfigParser
        from collections import OrderedDict
        parser = ConfigParser(dict_type=OrderedDict)
        multiThreads = confObj.getConfigValue("collect_traffic_use_threads") == "true"
        if not multiThreads:
            parser.add_section("general")
            parser.set("general", "server_multithreaded", str(multiThreads))
        cmdTraffic = confObj.getCompositeConfigValue("collect_traffic", "asynchronous")
        if cmdTraffic:
            parser.add_section("command line")
            parser.set("command line", "intercepts", ",".join(cmdTraffic))
            asynchronous = confObj.getConfigValue("collect_traffic").get("asynchronous", [])
            for cmd in cmdTraffic:
                env = confObj.getConfigValue("collect_traffic_environment").get(cmd)
                cmdAsync = cmd in asynchronous
                if env or cmdAsync:
                    parser.add_section(cmd)
                    if env:
                        parser.set(cmd, "environment", ",".join(env))
                    if cmdAsync:
                        parser.set(cmd, "asynchronous", str(cmdAsync))

        envVars = confObj.getConfigValue("collect_traffic_environment").get("default")
        if envVars:
            parser.set("command line", "environment", ",".join(envVars))

        pyTraffic = confObj.getConfigValue("collect_traffic_python")
        if pyTraffic:
            parser.add_section("python")
            ignore_callers = confObj.getConfigValue("collect_traffic_python_ignore_callers")
            parser.set("python", "intercepts", ",".join(pyTraffic))
            if ignore_callers:
                parser.set("python", "ignore_callers", ",".join(ignore_callers))

        if len(parser.sections()) > 0:  # don't write empty files
            print("Wrote file at", newFile)
            parser.write(open(newFile, "w"))

    def setUpApplication(self, app):
        newFile = os.path.join(app.getDirectory(), "capturemockrc." + app.name + app.versionSuffix())
        print("Converting", repr(app))
        self.convert(app, newFile)

    def __call__(self, test):
        self.checkTest(test)

    def setUpSuite(self, suite):
        self.checkTest(suite)

    def checkTest(self, test):
        configFile = test.getFileName("config")
        if configFile and test.parent:
            print("Converting", repr(test))
            newFile = os.path.join(test.getDirectory(), "capturemockrc." + test.app.name + test.app.versionSuffix())
            self.convert(test, newFile)



#!/usr/bin/env python3


import os
import sys
from optparse import OptionParser


def fixSysPath(fileName):
    install_root = os.path.dirname(os.path.dirname(fileName))
    # We pick up the basic libraries.
    # or a "generic" directory containing the TextTest core with local modules in the root
    for subdir in ["", "lib", "site/lib", "generic"]:
        libDir = os.path.abspath(os.path.join(install_root, subdir))
        if os.path.isdir(libDir):
            sys.path.insert(0, libDir)


if __name__ == "__main__":
    fixSysPath(os.path.abspath(__file__))
    from texttestlib import plugins, default
    parser = OptionParser("usage: %prog [options] filter1 filter2 ...")
    parser.add_option("-m", "--module",
                      help="also import module MODULE", metavar="MODULE")
    parser.add_option("-u", "--unordered", action="store_true",
                      help='Use unordered filter instead of standard one')
    parser.add_option("-t", "--testrelpath",
                      help="use test relative path RELPATH", metavar="RELPATH")
    (options, args) = parser.parse_args()
    if options.module:
        exec("import " + options.module)
    allPaths = plugins.findDataPaths(["logging.console"], dataDirName="log", includePersonal=True)
    plugins.configureLogging(allPaths[-1])  # Won't have any effect if we've already got a log file
    if options.unordered:
        runDepFilter = default.rundependent.UnorderedTextFilter(args, options.testrelpath)
    else:
        runDepFilter = default.rundependent.RunDependentTextFilter(args, options.testrelpath)
    runDepFilter.filterFile(sys.stdin, sys.stdout)



import urllib.request
import base64
import json
import re

def _makeURL(location, bugText):
    return location + "/browse/" + bugText

def _getEntry(key, fields, secondaryKey="name", secondaryEntryList=False):
    entry = fields[key]
    if secondaryEntryList:
        result = []
        for item in entry:
            result.append(item[secondaryKey])
        if not result:
            entry = "N/A"
        else:
            entry = ", ".join(map(_encodeString, result))
    elif secondaryKey:
        entry = _encodeString(entry[secondaryKey])

    return key.capitalize() + ": " + _encodeString(entry)

def _parseReply(json_dict, location):
    bugId = _encodeString(json_dict["key"])
    fields = json_dict["fields"]

    ruler = "*" * 50
    message_rows = [ "\n" + ruler,
                     bugId,
                     _getEntry("summary", fields, None),
                     _getEntry("status", fields),
                     _getEntry("assignee", fields),
                     _getEntry("reporter", fields),
                     _getEntry(key="components",
                               fields=fields,
                               secondaryEntryList=True),
                     _getEntry("created", fields, None),
                     _getEntry("updated", fields, None),
                     _getEntry("priority", fields),
                     ruler,
                     "\nView bug " + bugId + " using Jira URL=" \
                     + _makeURL(location, str(bugId)) + "\n",
                     _encodeString(fields["description"]) + "\n"]

    status = _encodeString(fields["status"]["name"])

    message = "\n".join(message_rows)
    isResolved = fields["resolution"]
    return status, message, isResolved, bugId

# For python2 this function contains locale related encoding that didn't work
# in python3. Bugs related to encoding is probably related to this.
def _encodeString(value):
    # Get given Windows line endings but Python doesn't use them internally.
    return value.replace("\r", "")

def _getJson(query_url, username, password):
    # Setting up password managers and using openers as described in the
    # documentation for urllib did not work.
    credentials = ('%s:%s' % (username, password))
    encoded_credentials = base64.b64encode(credentials.encode('ascii'))
    request = urllib.request.Request(query_url)
    request.add_header('Authorization', 'Basic %s' % encoded_credentials.decode('ascii'))
    response = urllib.request.urlopen(request)
    response_text = response.read()
    return json.loads(response_text.decode())

def _handleHTTPError(http_error, location, rest_url):
    message = ""
    internal_error = False
    if http_error.getcode() == 401:
        message = "Authentication failure for '" + location + "':\n" + \
            str(http_error) + \
            "\nPossibly issues with Jira user name and/or password.\n"
        return message, False
    elif http_error.getcode() == 404:
        try:
            json_dict = json.loads(http_error.read())
            error_key = "errorMessages"
            if error_key in json_dict:
                message = "Message from Jira: " + " ".join(json_dict[error_key])
                internal_error = True
        except:
            message = str(http_error) + \
                ("\nIf 'bug_system_location' is correct and the Jira issue exists,"
                 " this might indicate that the Jira REST API changed.")
    if not message:
        message = str(http_error)
    return "Failure while accessing '" + rest_url + "': \n" + message + "\n", internal_error


def findBugInfo(bugId, location, username, password):
    fields_to_fetch = ["assignee",
                       "components",
                       "created",
                       "description",
                       "priority",
                       "reporter",
                       "resolution",
                       "status",
                       "updated",
                       "summary"]
    rest_url = location + "/rest/api/2/issue/" + bugId \
        + "?fields=" + ",".join(fields_to_fetch)

    jira_error_label = "JIRA ERROR"
    try:
        json_dict = _getJson(rest_url, username, password)
    except urllib.request.HTTPError as h:
        message, internal_error = _handleHTTPError(h, location, rest_url)
        return jira_error_label, message, internal_error, bugId
    except Exception as e:
        message = "Exception while accessing '" + rest_url + "':\n" + str(e)  + "\n"
        return jira_error_label, message, False, bugId

    try:
        return _parseReply(json_dict, location)
    except Exception as e:
        message = "Failed to parse reply from '" + rest_url + "':\n" + str(e) + "\n"
        return "PARSE ERROR", message, False, bugId


# Used by Jenkins plugin
def getBugsFromText(text, location):
    bugRegex = re.compile("[A-Z]{2,}-[0-9]+")
    bugs = []
    for match in bugRegex.finditer(text):
        bugText = match.group(0)
        bugs.append((bugText, _makeURL(location, bugText)))
    return bugs



#!/usr/bin/env python

from gi.repository import Gtk


class EntryCompletionManager:
    def __init__(self):
        self.completions = Gtk.ListStore(str)
        self.entries = []
        self.enabled = False
        self.useContainsFunction = False
        self.inlineCompletions = False

    def start(self, matching, inline, completions):
        self.enabled = True
        self.useContainsFunction = matching == 2
        self.inlineCompletions = inline

        for completion in completions:
            self.addTextCompletion(completion)

    def register(self, entry):
        if self.enabled:
            completion = Gtk.EntryCompletion()
            completion.set_model(self.completions)
            if self.inlineCompletions:
                completion.set_inline_completion(True)
            completion.set_text_column(0)
            if self.useContainsFunction:  # Matching on start is default for Gtk.EntryCompletion
                completion.set_match_func(self.containsMatchFunction)

            self.addCompletion(entry)
            entry.set_completion(completion)
            entry.connect('activate', self.addCompletion)
            self.entries.append(entry)

    def addCompletion(self, entry):
        self.addTextCompletion(entry.get_text())

    def addTextCompletion(self, text):
        if self.enabled and text and text not in [row[0] for row in self.completions]:
            self.completions.prepend([text])

    def collectCompletions(self):
        if self.enabled:
            for entry in self.entries:
                self.addCompletion(entry)

    # Return true for any completion containing the key
    def containsMatchFunction(self, dummy, key_string, iter):
        value = self.completions.get_value(iter, 0)
        return value and value.lower().find(key_string) != -1


manager = EntryCompletionManager()



#!/usr/bin/env python

# Wrapper for Xvfb. Works on UNIX only. Main points are as follows :
# - Provide a unique display ID by using our own process ID
# - Get Xvfb to ignore job control signals properly
# - Handle Xvfb's weird mechanism whereby it sends SIGUSR1 to its parent process
#   every time a process disconnects. If the parent process is TextTest there is no
#   way to tell these signals from ordinary job control
# - Clean up after Xvfb as it leaks lock files from time to time.

import os
import signal
import sys
import subprocess
from socket import gethostname

MAX_DISPLAY = 32768
Xvfb_ready = False


class ConnectionComplete(BaseException):
    pass


class ConnectionTimeout(BaseException):
    pass


def setReadyFlag(self, *args):  # pragma: no cover - only here to deal with pathological and probably impossible race condition
    global Xvfb_ready
    Xvfb_ready = True


def connectionComplete(self, *args):
    raise ConnectionComplete()


def connectionFailed(self, *args):
    raise ConnectionTimeout()


def ignoreSignals():
    for signum in [signal.SIGUSR1, signal.SIGUSR2, signal.SIGXCPU]:
        signal.signal(signum, signal.SIG_IGN)


def getDisplayNumber():
    # We use the device of making the display number match our process ID (mod 32768)!
    # And we hope that works :) Should prevent clashes with others using the same strategy anyway
    # Display numbers up to 32768 seem to be allowed, which is less than most process IDs on systems I've observed...
    return str(os.getpid() % MAX_DISPLAY)


def getLockFiles(num):
    lockFile = "/tmp/.X" + num + "-lock"
    xFile = "/tmp/.X11-unix/X" + num
    return [lockFile, xFile]


def cleanLeakedLockFiles(displayNum):
    # Xvfb sometimes leaves lock files lying around, clean up
    for lockFile in getLockFiles(displayNum):
        if os.path.isfile(lockFile):
            try:
                os.remove(lockFile)
            except:  # pragma: no cover - pathological case of ending up in race condition with Xvfb
                pass


def writeAndWait(text, proc, displayNum):
    ignoreSignals()
    sys.stdout.write(text + "\n")
    sys.stdout.flush()
    proc.wait()
    cleanLeakedLockFiles(displayNum)


def runXvfb(logDir, extraArgs):
    ignoreSignals()
    signal.signal(signal.SIGUSR1, setReadyFlag)
    displayNum = getDisplayNumber()
    logFile = os.path.join(logDir, "Xvfb." + displayNum + "." + gethostname().split(".")
                           [0] + ".diag") if logDir != os.devnull else os.devnull
    startArgs = ["Xvfb", "-ac"] + extraArgs + ["-audit", "2", ":" + displayNum]
    proc = subprocess.Popen(startArgs, preexec_fn=ignoreSignals,
                            stdout=open(logFile, "w"), stderr=subprocess.STDOUT, stdin=open(os.devnull))
    try:
        signal.signal(signal.SIGUSR1, connectionComplete)
        signal.signal(signal.SIGALRM, connectionFailed)
        if not Xvfb_ready:
            signal.alarm(int(os.getenv("TEXTTEST_XVFB_WAIT", 15)))  # Time to wait for Xvfb to set up connections
            signal.pause()
    except ConnectionTimeout:
        # Kill it and tell TextTest we timed out. It will then start a new startXvfb.py process, with a new process ID
        # that will hopefully work better
        os.kill(proc.pid, signal.SIGTERM)
        return writeAndWait("Time Out!", proc, displayNum)
    except ConnectionComplete:
        signal.alarm(0)  # cancel any alarms that were previously set up!

    writeAndWait(displayNum + "," + str(proc.pid), proc, displayNum)


if __name__ == "__main__":
    runXvfb(sys.argv[1], sys.argv[2:])



#!/usr/bin/env python

import os
import sys
import subprocess
from socket import gethostname


def runWindowManager(logDir, wmExecutable, displayName, extraArgs):
    if not (logDir and wmExecutable and displayName):
        sys.stdout.write("Cannot find required arguments for logging directory, X display name and WM executable.")
        return 1
    command = [wmExecutable] + extraArgs
    logFileName = os.path.basename(os.path.normpath(wmExecutable)) + "." + \
        displayName.lstrip(":") + "." + gethostname().split(".")[0] + ".diag"
    logFilePath = os.path.join(logDir, logFileName) if logDir != os.devnull else os.devnull

    subEnv = dict(os.environ, DISPLAY=displayName)
    proc = subprocess.Popen(command, env=subEnv, stdout=open(logFilePath, "w"),
                            stderr=subprocess.STDOUT, stdin=open(os.devnull))
    sys.stdout.write(str(proc.pid) + "\n")
    sys.stdout.flush()
    return proc.wait()


if __name__ == "__main__":
    runWindowManager(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4:])




import logging
import os
from collections import OrderedDict
from .batchutils import getBatchRunName
from string import Template
from locale import getpreferredencoding
from texttestlib import plugins
from texttestlib.default.performance import getPerformance
from xml.sax.saxutils import escape


class JUnitResponder(plugins.Responder):
    """Respond to test results and write out results in format suitable for JUnit
    report writer. Only does anything if the app has batch_junit_format:true in its config file """

    def __init__(self, optionMap, *args):
        plugins.Responder.__init__(self)
        self.runId = getBatchRunName(optionMap)
        self.allApps = OrderedDict()
        self.appData = OrderedDict()

    def useJUnitFormat(self, app):
        return app.getBatchConfigValue("batch_junit_format") == "true"

    def notifyComplete(self, test):
        if not self.useJUnitFormat(test.app):
            return
        if test.app not in self.appData:
            self._addApplication(test)
        self.appData[test.app].storeResult(test)

    def notifyAllComplete(self):
        # allApps is {appname : [app]}
        for appList in list(self.allApps.values()):
            # appData is {app : data}
            for app in appList:
                if self.useJUnitFormat(app):
                    data = self.appData[app]
                    ReportWriter(self.runId).writeResults(app, data)

    def _addApplication(self, test):
        app = test.app
        self.appData[app] = JUnitApplicationData()
        self.allApps.setdefault(app.name, []).append(app)


class JUnitApplicationData:
    """Data class to store test results in a format convenient for conversion to 
    JUnit report format """

    def __init__(self):
        self.testResults = {}

    def storeResult(self, test):
        perfStem = test.getConfigValue("default_performance_stem")
        perfFile = test.makeTmpFileName(perfStem)
        t = getPerformance(perfFile) if os.path.isfile(perfFile) else 0
        result = dict(full_test_name=self._fullTestName(test),
                      test_name=test.name,
                      suite_name=self._suiteName(test),
                      encoding=getpreferredencoding(),
                      time=str(t))
        if not test.state.hasResults():
            self._error(test, result)
        elif test.state.hasSucceeded():
            self._success(result)
        else:
            self._failure(test, result)

        self.testResults[test.getRelPath().replace(os.sep, ".")] = result

    def getResults(self):
        return self.testResults

    def _suiteName(self, test):
        fullName = self._fullTestName(test)
        return ".".join(fullName.split(".")[:-1])

    def _fullTestName(self, test):
        relpath = test.getRelPath()
        return test.app.fullName() + "." + relpath.replace(os.sep, ".")

    def _error(self, test, result):
        result["error"] = True
        result["success"] = False
        result["failure"] = False
        result["short_message"] = self._shortMessage(test)
        result["long_message"] = self._longMessage(test)

    def _success(self, result):
        result["error"] = False
        result["success"] = True
        result["failure"] = False

    def _failure(self, test, result):
        result["error"] = False
        result["success"] = False
        result["failure"] = True
        result["short_message"] = self._shortMessage(test)
        result["long_message"] = self._longMessage(test)

    def _shortMessage(self, test):
        return escape(test.state.getTypeBreakdown()[1])

    def _longMessage(self, test):
        message = test.state.freeText.replace("]]>", "END_MARKER")
        return self._char_filter(message)

    @classmethod
    def _char_filter(cls, text):
        """
        Replace char with `"\\x%02x" % ord(char)` if char not allowed in CDATA.
        """
        return "".join(("\\x%02x" % ord(char), char)[cls._allowed(ord(char))] for char in text)

    @staticmethod
    def _allowed(num):
        """
        See http://www.w3.org/TR/REC-xml/#NT-Char
        """
        if num < 0x20:
            return num in (0x9, 0xA, 0xD)
        if num <= 0xD7FF:
            return True
        if num < 0xE000:
            return False
        if num <= 0xFFFD:
            return True
        if num < 0x10000:
            return False
        if num <= 0x10FFFF:
            return True
        return False


failure_template = """\
<?xml version="1.0" encoding="$encoding"?>
<testsuite name="$full_test_name" failures="1" tests="1" time="$time" errors="0">
  <properties/>
  <testcase name="$test_name" time="$time" classname="$suite_name">
    <failure type="differences" message="$short_message">
    <![CDATA[
$long_message
]]>
    </failure>
  </testcase>
</testsuite>
"""

error_template = """\
<?xml version="1.0" encoding="$encoding"?>
<testsuite name="$full_test_name" failures="0" tests="1" time="$time" errors="1">
  <properties/>
  <testcase name="$test_name" time="$time" classname="$suite_name">
    <error type="none" message="$short_message">
    <![CDATA[
$long_message
]]>
    </error>
  </testcase>
</testsuite>
"""

success_template = """\
<?xml version="1.0" encoding="$encoding"?>
<testsuite name="$full_test_name" failures="0" tests="1" time="$time" errors="0">
  <properties/>
  <testcase name="$test_name" time="$time" classname="$suite_name"/>
</testsuite>
"""


class ReportWriter:
    def __init__(self, runId):
        self.runId = runId
        self.diag = logging.getLogger("JUnit Report Writer")

    def writeResults(self, app, appData):
        self.diag.info("writing results in junit format for app " + app.fullName())
        appResultsDir = self._createResultsDir(app)
        for testName, result in list(appData.getResults().items()):
            if result["success"]:
                text = Template(success_template).substitute(result)
            elif result["error"]:
                text = Template(error_template).substitute(result)
            else:
                text = Template(failure_template).substitute(result)
            testFileName = os.path.join(appResultsDir, testName + ".xml")
            self._writeTestResult(testFileName, text)

    def _writeTestResult(self, testFileName, text):
        testFile = open(testFileName, "w")
        testFile.write(text)
        testFile.close()

    def _createResultsDir(self, app):
        resultsDir = self.userDefinedFolder(app)
        if (resultsDir is None or resultsDir.strip() == ""):
            resultsDir = os.path.join(app.writeDirectory, "junitreport")

        if not os.path.exists(resultsDir):
            os.mkdir(resultsDir)
        appResultsDir = os.path.join(resultsDir, app.name + app.versionSuffix())
        if not os.path.exists(appResultsDir):
            os.mkdir(appResultsDir)
        return appResultsDir

    def userDefinedFolder(self, app):
        return app.getBatchConfigValue("batch_junit_folder")




import os
from . import gridqueuesystem

# Used by the master to submit, monitor and delete jobs...


class QueueSystem(gridqueuesystem.QueueSystem):
    submitProg = "bsub"
    def getSubmitCmdArgs(self, submissionRules, commandArgs=[], slaveEnv={}):
        bsubArgs = ["bsub", "-J", submissionRules.getJobName()]
        if submissionRules.processesNeeded != 1:
            bsubArgs += ["-n", str(submissionRules.processesNeeded)]
        queue = submissionRules.findQueue()
        if queue:
            bsubArgs += ["-q", queue]
        resource = self.getResourceArg(submissionRules)
        if len(resource):
            bsubArgs += ["-R", resource]
        machines = submissionRules.findMachineList()
        if len(machines):
            bsubArgs += ["-m", " ".join(machines)]
        bsubArgs += ["-u", "nobody", "-o", os.devnull, "-e", os.devnull]
        return self.addExtraAndCommand(bsubArgs, submissionRules, commandArgs)

    def getSlaveVarsToBlock(self):
        """Make sure we clear out the master scripts so the slave doesn't use them too,
        otherwise just use the environment as is.

        If we're being run via SSH, don't pass this on to the slave jobs
        This has been known to trip up shell starter scripts, e.g. on SuSE 10
        making them believe that the SGE job is an SSH login and setting things wrongly
        as a result.

        LS_COLORS has also been shown to be problematic as older version of tcsh fail hard
        if given newer instructions they don't understand there.
        """
        return ["USECASE_REPLAY_SCRIPT", "USECASE_RECORD_SCRIPT", "SSH_TTY", "LS_COLORS"]

    def findSubmitError(self, stderr):
        for errorMessage in stderr.splitlines():
            if self.isRealError(errorMessage):
                return errorMessage

    def isRealError(self, errorMessage):
        if not errorMessage:
            return 0
        okStrings = ["still trying", "Waiting for dispatch", "Job is finished"]
        for okStr in okStrings:
            if errorMessage.find(okStr) != -1:
                return 0
        return 1

    def _getJobFailureInfo(self, jobId):
        resultOutput = os.popen("bjobs -a -l " + jobId + " 2>&1").read()
        if resultOutput.find("is not found") != -1:
            return "LSF lost job:" + jobId
        else:
            return resultOutput

    def supportsPolling(self):
        # This feature was added to the SGE handling when I no longer had access to an LSF cluster
        return False

    def killJob(self, jobId):
        resultOutput = os.popen("bkill -s USR1 " + jobId + " 2>&1").read()
        return resultOutput.find("is being terminated") != -1 or resultOutput.find("is being signaled") != -1

    def getJobId(self, line):
        word = line.split()[1]
        return word[1:-1]

    def findJobId(self, stdout):
        for line in stdout.splitlines():
            if line.find("is submitted") != -1:
                return self.getJobId(line)
            else:
                print("Unexpected output from bsub :", line.strip())
        return ""  # pragma : no cover, should never happen...

    def getResourceArg(self, submissionRules):
        resourceList = submissionRules.findResourceList()
        if len(resourceList) == 0:
            return ""
        selectResources = []
        others = []
        for resource in resourceList:
            if resource.find("rusage[") != -1 or resource.find("order[") != -1 or \
               resource.find("span[") != -1 or resource.find("same[") != -1:
                others.append(resource)
            else:
                selectResources.append(resource)
        if len(selectResources) == 0:
            return " ".join(others)
        else:
            return self.getSelectResourceArg(selectResources) + " " + " ".join(others)

    def getSelectResourceArg(self, resourceList):
        if len(resourceList) == 1:
            return self.formatResource(resourceList[0])
        else:
            resource = "(" + self.formatResource(resourceList[0]) + ")"
            for res in resourceList[1:]:
                resource += " && (" + self.formatResource(res) + ")"
            return resource

    def formatResource(self, res):
        if res.find("==") == -1 and res.find("!=") == -1 and res.find("<=") == -1 and \
           res.find(">=") == -1 and res.find("=") != -1:
            return res.replace("=", "==")
        else:
            return res

# Used by the slave for getting performance info


class MachineInfo:
    def findActualMachines(self, machineOrGroup):
        machines = []
        for line in os.popen("bhosts " + machineOrGroup + " 2>&1"):
            if not line.startswith("HOST_NAME"):
                machines.append(line.split()[0].split(".")[0])
        return machines

    def findResourceMachines(self, resource):
        machines = []
        for line in os.popen("bhosts -w -R '" + resource + "' 2>&1"):
            if not line.startswith("HOST_NAME"):
                machines.append(line.split()[0].split(".")[0])
        return machines

    def findRunningJobs(self, machine):
        jobs = []
        for line in os.popen("bjobs -m " + machine + " -u all -w 2>&1 | grep RUN"):
            fields = line.split()
            jobId = fields[0]
            user = fields[1]
            jobName = fields[6]
            jobs.append((user, jobId, jobName))
        return jobs

# Interpret what the limit signals mean...


def getUserSignalKillInfo(userSignalNumber, explicitKillMethod):
    if userSignalNumber == "2":
        return "RUNLIMIT", "exceeded maximum wallclock time allowed by LSF (RUNLIMIT parameter)"
    else:
        return explicitKillMethod()

# Need to get all hosts for parallel


def getExecutionMachines():
    if "LSB_HOSTS" in os.environ:
        hosts = os.environ["LSB_HOSTS"].split()
        return [host.split(".")[0] for host in hosts]
    else:
        from texttestlib.plugins import gethostname
        return [gethostname()]



# Code to generate HTML report of historical information. This report generated
# either via the -coll flag, or via -s 'batch.GenerateHistoricalReport <batchid>'

import os
import time
import html
import sys
import logging
import locale
from texttestlib.default.batch import HTMLgen, HTMLcolors, jenkinschanges
from texttestlib import plugins
from collections import OrderedDict
from glob import glob
from pprint import pformat
from datetime import datetime, timedelta
from .batchutils import convertToUrl, getEnvironmentFromRunFiles
HTMLgen.PRINTECHO = 0


def parseState(state):
    '''Separate brief text and host list from succeeded_runs file, e.g.
    21Jan2006 after 7 reruns placerville, placerville
    Date field is not part of state.'''
    hosts = state.split(", ")
    possibleBrief = hosts[0].rsplit(" ", 1)
    if len(possibleBrief) > 1:
        # possibleBrief is actually [brief, host0]
        hosts[0] = possibleBrief[1]
        return possibleBrief[0] + " ", hosts
    else:
        return "", hosts


def compactHostRepr(hosts):
    """Convert list ['a', 'a', 'b'] -> 'a*2, b'"""
    hostCount = dict((host, hosts.count(host)) for host in set(hosts))
    hostCountStrings = \
        (('*'.join([host, str(hostCount[host])])
          if hostCount[host] > 1 else host
          for host in hostCount))
    return ', '.join(hostCountStrings)


def getWeekDay(tag):
    return plugins.weekdays[time.strptime(tag.split("_")[0], "%d%b%Y")[6]]


class ColourFinder:
    def __init__(self, getConfigValue):
        self.getConfigValue = getConfigValue

    def find(self, title):
        colourName = self.getConfigValue("historical_report_colours", title)
        return self.htmlColour(colourName)

    def htmlColour(self, colourName):
        if colourName and not colourName.startswith("#"):
            return getattr(HTMLcolors, colourName.upper())
        else:
            return colourName


def getDisplayText(tag):
    displayText = "_".join(tag.split("_")[1:])
    if displayText:
        return displayText
    else:
        return tag


def makeTestId(version, extraVersion, testName):
    testId = version
    if extraVersion:
        testId += "." + extraVersion
    testId += testName
    return testId


class TitleWithDateStamp:
    def __init__(self, title):
        self.title = title + " (generated by " + os.getenv("USER", os.getenv("USERNAME", "unknown")) + " at "

    def __str__(self):
        return self.title + plugins.localtime(format="%d%b%H:%M") + ")"


class GenerateWebPages(object):
    def __init__(self, getConfigValue, pageDir, resourceNames,
                 pageTitle, pageSubTitles, pageVersion, extraVersions, descriptionInfo):
        self.pageTitle = pageTitle
        self.pageSubTitles = pageSubTitles
        self.pageVersion = pageVersion
        self.extraVersions = extraVersions
        self.pageDir = pageDir
        self.pagesOverview = OrderedDict()
        self.pagesDetails = OrderedDict()
        self.getConfigValue = getConfigValue
        self.resourceNames = resourceNames
        self.descriptionInfo = descriptionInfo
        self.diag = logging.getLogger("GenerateWebPages")

    def makeSelectors(self, subPageNames, tags=[]):
        allSelectors = []
        firstSubPageName = self.getConfigValue("historical_report_subpages", "default")[0]
        for subPageName in subPageNames:
            if subPageName == firstSubPageName:
                suffix = ""
            else:
                suffix = "_" + subPageName.lower()
            allSelectors.append(Selector(subPageName, suffix, self.getConfigValue, tags))
        return allSelectors

    def removeUnused(self, unused, tagData):
        successTags = {}
        for tag in unused:
            for fn in tagData.get(tag):
                if os.path.basename(fn).startswith("teststate_"):
                    os.remove(fn)
                else:
                    successTags.setdefault(fn, []).append(tag)
        for fn, tagsToRemove in list(successTags.items()):
            linesToKeep = []
            with open(fn) as readFile:
                for line in readFile:
                    tag = line.strip().split()[0]
                    if tag not in tagsToRemove:
                        linesToKeep.append(line)

            with open(fn, "w") as writeFile:
                for line in linesToKeep:
                    writeFile.write(line)

    def generate(self, repositoryDirs, subPageNames, archiveUnused):
        minorVersionHeader = HTMLgen.Container()
        allMonthSelectors = set()
        latestMonth = None
        pageToGraphs = {}
        for version, repositoryDirInfo in list(repositoryDirs.items()):
            self.diag.info("Generating " + version)
            tagData, stateFiles, successFiles = self.findTestStateFilesAndTags(repositoryDirInfo)
            if len(stateFiles) > 0 or len(successFiles) > 0:
                tags = list(tagData.keys())
                tags.sort(key=self.tagSortKey)
                selectors = self.makeSelectors(subPageNames, tags)
                monthSelectors = SelectorByMonth.makeInstances(tags)
                allMonthSelectors.update(monthSelectors)
                allSelectors = selectors + list(reversed(monthSelectors))
                # If we already have month pages, we only regenerate the current one
                if len(self.getExistingMonthPages()) == 0:
                    selectors = allSelectors
                else:
                    currLatestMonthSel = monthSelectors[-1]
                    if latestMonth is None or currLatestMonthSel.linkName == latestMonth:
                        selectors.append(monthSelectors[-1])
                        latestMonth = currLatestMonthSel.linkName
                    selectedTags = set()
                    unusedTags = set(tags)
                    for selector in selectors:
                        currTags = set(selector.selectedTags)
                        selectedTags.update(currTags)
                        if archiveUnused:
                            unusedTags.difference_update(currTags)
                    tags = [t for t in tags if t in selectedTags]
                    if archiveUnused and unusedTags:
                        plugins.log.info(
                            "Automatic repository cleaning will now remove old data for the following runs:")
                        for tag in sorted(unusedTags, key=self.tagSortKey):
                            plugins.log.info("- " + tag)
                        plugins.log.info(
                            "(To disable automatic repository cleaning in future, please run with the --manualarchive flag when collating the HTML report.)")
                        self.removeUnused(unusedTags, tagData)

                loggedTests = OrderedDict()
                categoryHandlers = {}
                self.diag.info("Processing " + str(len(stateFiles)) + " teststate files")
                relevantFiles = 0
                for stateFile, repository in stateFiles:
                    tag = self.getTagFromFile(stateFile)
                    if len(tags) == 0 or tag in tags:
                        relevantFiles += 1
                        testId, state, extraVersion = self.processTestStateFile(stateFile, repository)
                        loggedTests.setdefault(extraVersion, OrderedDict()).setdefault(
                            testId, OrderedDict())[tag] = state
                        categoryHandlers.setdefault(tag, CategoryHandler()).registerInCategory(
                            testId, state.category, extraVersion, state)
                        if relevantFiles % 100 == 0:
                            self.diag.info("- Processed " + str(relevantFiles) + " files with matching tags so far")
                self.diag.info("Processed " + str(relevantFiles) + " relevant teststate files")
                self.diag.info("Processing " + str(len(successFiles)) + " success files")
                for successFile, repository in successFiles:
                    testId = self.getTestIdentifier(successFile, repository)
                    extraVersion = self.findExtraVersion(repository)
                    with open(successFile) as f:
                        fileTags = set()
                        for line in f:
                            parts = line.strip().split(" ", 1)
                            if len(parts) != 2:
                                continue
                            tag, text = parts
                            if tag in fileTags:
                                sys.stderr.write("WARNING: more than one result present for tag '" +
                                                 tag + "' in file " + successFile + "!\n")
                                sys.stderr.write("Ignoring later ones\n")
                                continue

                            fileTags.add(tag)
                            if len(tags) == 0 or tag in tags:
                                loggedTests.setdefault(extraVersion, OrderedDict()).setdefault(
                                    testId, OrderedDict())[tag] = text
                                categoryHandlers.setdefault(tag, CategoryHandler()).registerInCategory(
                                    testId, "success", extraVersion, text)
                self.diag.info("Processed " + str(len(successFiles)) + " success files")
                versionToShow = self.removePageVersion(version)
                hasData = False
                for sel in selectors:
                    filePath = self.getPageFilePath(sel)
                    if filePath in self.pagesOverview:
                        page, pageColours = self.pagesOverview[filePath]
                    else:
                        page = self.createPage()
                        pageColours = set()
                        self.pagesOverview[filePath] = page, pageColours

                    tableHeader = self.getTableHeader(version, repositoryDirs)
                    heading = self.getHeading(versionToShow)
                    hasNewData, graphLink, tableColours = self.addTable(page, self.resourceNames, categoryHandlers, version,
                                                                        loggedTests, sel, tableHeader, filePath, heading, repositoryDirInfo)
                    hasData |= hasNewData
                    pageColours.update(tableColours)
                    if graphLink:
                        pageToGraphs.setdefault(page, []).append(graphLink)

                if hasData and versionToShow:
                    link = HTMLgen.Href("#" + version, versionToShow)
                    minorVersionHeader.append(link)

                # put them in reverse order, most relevant first
                linkFromDetailsToOverview = [sel.getLinkInfo(self.pageVersion) for sel in allSelectors]
                for tag in tags:
                    details = self.pagesDetails.setdefault(tag, TestDetails(tag, self.pageTitle, self.pageSubTitles))
                    details.addVersionSection(version, categoryHandlers[tag], linkFromDetailsToOverview)

        selContainer = HTMLgen.Container()
        selectors = self.makeSelectors(subPageNames)
        for sel in selectors:
            target, linkName = sel.getLinkInfo(self.pageVersion)
            selContainer.append(HTMLgen.Href(target, linkName))

        monthContainer = HTMLgen.Container()
        if len(allMonthSelectors) == 1:
            # Don't want just one month, no navigation possible
            prevMonth = list(allMonthSelectors)[0].getPreviousMonthSelector()
            allMonthSelectors.add(prevMonth)

        for sel in sorted(allMonthSelectors, key=lambda s: s.sortKey()):
            target, linkName = sel.getLinkInfo(self.pageVersion)
            monthContainer.append(HTMLgen.Href(target, linkName))

        for page, pageColours in list(self.pagesOverview.values()):
            if len(monthContainer.contents) > 0:
                page.prepend(HTMLgen.Heading(2, monthContainer, align='center'))
            graphs = pageToGraphs.get(page)
            page.prepend(HTMLgen.Heading(2, selContainer, align='center'))
            if minorVersionHeader.contents:
                if not graphs is None and len(graphs) > 1:
                    page.prepend(HTMLgen.Heading(1, *graphs, align='center'))
                page.prepend(HTMLgen.Heading(1, minorVersionHeader, align='center'))
            creationDate = TitleWithDateStamp("").__str__().strip()
            page.prepend(HTMLgen.Paragraph(creationDate, align="center"))
            page.prepend(HTMLgen.Heading(1, self.getHeading(), align='center'))
            if len(pageColours) > 0:
                page.prepend(HTMLgen.BR())
                page.prepend(HTMLgen.BR())
                page.script = self.getFilterScripts(pageColours)

        self.writePages()

    def getFilterScripts(self, pageColours):
        finder = ColourFinder(self.getConfigValue)
        rowHeaderColour = finder.find("row_header_bg")
        successColour = finder.find("success_bg")
        # Always put green at the start, we often want to filter that
        sortedColours = sorted(pageColours, key=lambda c: (c != successColour, c))
        scriptCode = "var TEST_ROW_HEADER_COLOR = " + repr(rowHeaderColour) + ";\n" + \
                     "var Colors = " + repr(sortedColours) + ";"
        return [HTMLgen.Script(code=scriptCode),
                HTMLgen.Script(src="../javascript/jquery.js"),
                HTMLgen.Script(src="../javascript/filter.js"),
                HTMLgen.Script(src="../javascript/comment.js")]

    def getHeading(self, versionToShow=""):
        heading = "Test results for " + self.pageTitle
        if versionToShow:
            heading += "." + versionToShow
        return heading

    def getTableHeader(self, version, repositoryDirs):
        return version if len(repositoryDirs) > 1 else ""

    def getExistingMonthPages(self):
        return glob(os.path.join(self.pageDir, "test_" + self.pageVersion + "_all_???[0-9][0-9][0-9][0-9].html"))

    def tagSortKey(self, x):
        return self.getTagTimeInSeconds(x), plugins.padNumbersWithZeroes(x)

    def getTagFromFile(self, fileName):
        return os.path.basename(fileName).replace("teststate_", "")

    def findTestStateFilesAndTags(self, repositoryDirs):
        tagData, stateFiles, successFiles = {}, [], []
        for _, dir in repositoryDirs:
            self.diag.info("Looking for teststate files in " + dir)
            for root, _, files in sorted(os.walk(dir)):
                for file in files:
                    path = os.path.join(root, file)
                    if file.startswith("teststate_"):
                        tag = self.getTagFromFile(file)
                        stateFiles.append((path, dir))
                        tagData.setdefault(tag, []).append(path)
                    elif file.startswith("succeeded_"):
                        successFiles.append((path, dir))
                        with open(path) as f:
                            for line in f:
                                parts = line.split()
                                if parts:
                                    tag = parts[0]
                                    tagData.setdefault(tag, []).append(path)

            self.diag.info("Found " + str(len(stateFiles)) + " teststate files and " +
                           str(len(successFiles)) + " success files in " + dir)
        return tagData, stateFiles, successFiles

    def processTestStateFile(self, stateFile, repository):
        state = self.readState(stateFile)
        testId = self.getTestIdentifier(stateFile, repository)
        extraVersion = self.findExtraVersion(repository)
        return testId, state, extraVersion

    def findExtraVersion(self, repository):
        versions = os.path.basename(repository).split(".")
        for i in range(len(versions)):
            version = ".".join(versions[i:])
            if version in self.extraVersions:
                return version
        return ""

    @classmethod
    def readState(cls, stateFile):
        file = open(stateFile, "rb")
        try:
            state = plugins.getNewTestStateFromFile(file)
            if isinstance(state, plugins.TestState):
                return state
            else:
                return cls.readErrorState("Incorrect type for state object.")
        except Exception as e:
            if os.path.getsize(stateFile) > 0:
                return cls.readErrorState("Stack info follows:\n" + str(e))
            else:
                return plugins.Unrunnable("Results file was empty, probably the disk it resides on is full.", "Disk full?")

    @staticmethod
    def readErrorState(errMsg):
        freeText = "Failed to read results file, possibly deprecated format. " + errMsg
        return plugins.Unrunnable(freeText, "read error")

    def removePageVersion(self, version):
        leftVersions = []
        pageSubVersions = self.pageVersion.split(".")
        for subVersion in version.split("."):
            if not subVersion in pageSubVersions:
                leftVersions.append(subVersion)
        return ".".join(leftVersions)

    def getPageFilePath(self, selector):
        pageName = selector.getLinkInfo(self.pageVersion)[0]
        return os.path.join(self.pageDir, pageName)

    def createPage(self):
        style = "body,td {color: #000000;font-size: 11px;font-family: Helvetica;} th {color: #000000;font-size: 13px;font-family: Helvetica;}"
        title = "Test results for " + self.pageTitle
        return HTMLgen.SimpleDocument(title=title, style=style, xhtml=True, meta='<meta charset="' + locale.getpreferredencoding() + '">')

    def makeTableHeaderCell(self, tableHeader):
        container = HTMLgen.Container()
        container.append(HTMLgen.Name(tableHeader))
        container.append(HTMLgen.U(HTMLgen.Heading(1, tableHeader, align='center')))
        return HTMLgen.TD(container)

    def makeImageLink(self, graphFile):
        image = HTMLgen.Image(filename=graphFile, src=graphFile, height=100, width=150, border=0)
        return HTMLgen.Href(graphFile, image)

    def getTestTable(self, *args):
        return TestTable(*args)

    def addTable(self, page, resourceNames, categoryHandlers, version, loggedTests, selector, tableHeader, filePath, graphHeading, repositoryDirs):
        graphDirname, graphFileRef = self.getGraphFileParts(filePath, version)
        testTable = self.getTestTable(self.getConfigValue, resourceNames, self.descriptionInfo,
                                      selector.selectedTags, categoryHandlers, self.pageVersion, version, os.path.join(graphDirname, graphFileRef))
        table = testTable.generate(loggedTests, self.pageDir, repositoryDirs)
        if table:
            cells = []
            if tableHeader:
                page.append(HTMLgen.HR())
                cells.append(self.makeTableHeaderCell(tableHeader))

            graphLink = None
            fullPath = os.path.abspath(os.path.join(graphDirname, graphFileRef))
            if testTable.generateGraph(fullPath, graphHeading):
                graphLink = self.makeImageLink(graphFileRef.replace("\\", "/"))
                cells.append(HTMLgen.TD(graphLink))

            if len(cells):
                row = HTMLgen.TR(*cells)
                initialTable = HTMLgen.TableLite(align="center")
                initialTable.append(row)
                page.append(initialTable)

            extraVersions = list(loggedTests.keys())[1:]
            if len(extraVersions) > 0:
                page.append(testTable.generateExtraVersionLinks(extraVersions))

            page.append(table)
            return True, graphLink, testTable.usedColours
        else:
            return False, None, []

    def getGraphFileParts(self, filePath, version):
        dirname, local = os.path.split(filePath)
        versionSuffix = self.removePageVersion(version)
        if versionSuffix:
            versionSuffix = "." + versionSuffix
        return dirname, os.path.join("images", local[:-5] + versionSuffix + ".png")

    def writePages(self):
        plugins.log.info("Writing overview pages...")
        fileToUrl = self.getConfigValue("file_to_url", allSubKeys=True)
        for pageFile, (page, _) in list(self.pagesOverview.items()):
            page.write(pageFile)
            plugins.log.info("wrote: '" + plugins.relpath(pageFile, self.pageDir) + "'")
            if fileToUrl:
                url = convertToUrl(pageFile, fileToUrl)
                plugins.log.info("(URL is " + url + ")")
        plugins.log.info("Writing detail pages...")
        for tag, details in list(self.pagesDetails.items()):
            pageName = getDetailPageName(self.pageVersion, tag)
            details.write(os.path.join(self.pageDir, pageName))
            plugins.log.info("wrote: '" + pageName + "'")
        self.writeCommentListPage()

    def writeCommentListPage(self):
        filename = os.path.join(self.pageDir, "commentlist.html")
        plugins.log.info("Writing comment html page at " + filename + "...")
        doc = HTMLgen.SimpleDocument(title="All Comments", stylesheet="../javascript/commentliststyle.css", xhtml=True, meta='<meta charset="' + locale.getpreferredencoding() + '">')
        for scriptFile in ["jquery.js", "comment.js", "commentlist.js"]:
            doc.append(HTMLgen.Script(src="../javascript/" + scriptFile))
        doc.write(filename)

    def getTestIdentifier(self, stateFile, repository):
        dir = os.path.dirname(stateFile)
        return dir.replace(repository + os.sep, "").replace(os.sep, " ")

    def getTagTimeInSeconds(self, tag):
        timePart = tag.split("_")[0]
        return time.mktime(time.strptime(timePart, "%d%b%Y"))


class TestTable:
    def __init__(self, getConfigValue, resourceNames, descriptionInfo, tags, categoryHandlers, pageVersion, version, graphFilePath):
        self.getConfigValue = getConfigValue
        self.resourceNames = resourceNames
        self.colourFinder = ColourFinder(getConfigValue)
        self.descriptionInfo = descriptionInfo
        self.tags = tags
        self.categoryHandlers = categoryHandlers
        self.pageVersion = pageVersion
        self.version = version
        self.graphFilePath = graphFilePath  # For convenience in performance analyzer.
        self.usedColours = set()

    def generateGraph(self, fileName, heading):
        if len(self.tags) > 1:  # Don't bother with graphs when tests have only run once
            try:
                from .resultgraphs import GraphGenerator
            except Exception as e:
                sys.stderr.write("Not producing result graphs: " + str(e) + "\n")
                return False  # if matplotlib isn't installed or is too old

            data = self.getColourKeySummaryData()
            generator = GraphGenerator()
            generator.generateGraph(fileName, heading, data, self.colourFinder)
            return True
        else:
            return False

    def generate(self, loggedTests, pageDir, repositoryDirs):
        table = HTMLgen.TableLite(border=0, cellpadding=4, cellspacing=2, width="100%")
        table.append(self.generateTableHead(repositoryDirs))
        table.append(self.generateSummaries())
        if os.getenv("JENKINS_URL"):
            changeRow = self.generateJenkinsChanges(pageDir)
            if changeRow:
                table.append(changeRow)
        hasRows = False
        for extraVersion, testInfo in list(loggedTests.items()):
            currRows = []
            for test in sorted(testInfo.keys()):
                results = testInfo[test]
                rows = self.generateTestRows(test, extraVersion, results)
                if rows:
                    currRows += rows

            if len(currRows) == 0:
                continue
            else:
                hasRows = True

            # Add an extra line in the table only if there are several versions.
            if len(loggedTests) > 1:
                fullVersion = self.version
                if extraVersion:
                    fullVersion += "." + extraVersion
                table.append(self.generateExtraVersionHeader(fullVersion))
                table.append(self.generateSummaries(extraVersion))

            for row in currRows:
                table.append(row)

        if hasRows:
            table.append(HTMLgen.BR())
            return table

    def findJenkinsChanges(self, prevTag, tag, cacheDir):
        buildNumber = self.getJenkinsBuildNumber(tag)
        cacheFileOldName = os.path.join(cacheDir, buildNumber)
        cacheFile = os.path.join(cacheDir, tag)
        if os.path.isfile(cacheFileOldName):
            os.rename(cacheFileOldName, cacheFile)
        if os.path.isfile(cacheFile):
            return eval(open(cacheFile).read().strip())
        else:
            bugSystemData = self.getConfigValue("bug_system_location", allSubKeys=True)
            markedArtefacts = self.getConfigValue("batch_jenkins_marked_artefacts")
            fileFinder = self.getConfigValue("batch_jenkins_archive_file_pattern")
            prevBuildNumber = self.getJenkinsBuildNumber(prevTag) if prevTag else None
            if buildNumber.isdigit() and prevBuildNumber is not None:
                try:
                    allChanges = jenkinschanges.getChanges(
                        prevBuildNumber, buildNumber, bugSystemData, markedArtefacts, fileFinder, cacheDir)
                    plugins.ensureDirectoryExists(cacheDir)
                    with open(cacheFile, "w") as f:
                        f.write(pformat(allChanges) + "\n")
                    return allChanges
                except jenkinschanges.JobStillRunningException:
                    pass  # don't write to cache in this case
            return []

    def getJenkinsBuildNumber(self, tag):
        return tag.split(".")[-1]

    def generateJenkinsChanges(self, pageDir):
        cacheDir = os.path.join(os.path.dirname(pageDir), "jenkins_changes")
        bgColour = self.colourFinder.find("changes_header_bg")
        row = [HTMLgen.TD("Changes", bgcolor=bgColour)]
        hasData = False
        prevTag = None
        for tag in self.tags:
            allChanges = self.findJenkinsChanges(prevTag, tag, cacheDir)
            cont = HTMLgen.Container()
            aborted = False
            for i, (authorOrMessage, target, bugs) in enumerate(allChanges):
                if i:
                    cont.append(HTMLgen.BR())
                if target:
                    cont.append(HTMLgen.Href(target, authorOrMessage))
                else:
                    cont.append(HTMLgen.Font(authorOrMessage, color="red"))
                    aborted = "Aborted" in authorOrMessage
                for bugText, bugTarget in bugs:
                    cont.append(HTMLgen.Href(bugTarget, bugText))
                hasData = True
            row.append(HTMLgen.TD(cont, bgcolor=bgColour))
            if not aborted:
                prevTag = tag
        if hasData:
            return HTMLgen.TR(*row)

    def generateSummaries(self, extraVersion=None):
        bgColour = self.colourFinder.find("column_header_bg")
        row = [HTMLgen.TD("Summary", bgcolor=bgColour)]
        for tag in self.tags:
            categoryHandler = self.categoryHandlers[tag]
            detailPageName = getDetailPageName(self.pageVersion, tag)
            summary = categoryHandler.generateHTMLSummary(detailPageName + "#" + self.version, extraVersion)
            row.append(HTMLgen.TD(summary, bgcolor=bgColour))
        return HTMLgen.TR(*row)

    def getColourKeySummaryData(self):
        fullData = []
        for tag in self.tags:
            colourCount = OrderedDict()
            for colourKey in ["success", "knownbug", "slower", "faster", "performance",
                              "larger", "smaller", "memory", "failure", "incomplete"]:
                colourCount[colourKey] = 0
            categoryHandler = self.categoryHandlers[tag]
            basicData = categoryHandler.getSummaryData()[-1]
            for category, count in basicData:
                colourKey = self.getBackgroundColourKey(category)
                colourCount[colourKey] += count
            fullData.append((getDisplayText(tag), colourCount))
        return fullData

    def generateExtraVersionLinks(self, extraVersions):
        cont = HTMLgen.Container()
        for extra in extraVersions:
            fullName = self.version + "." + extra
            cont.append(HTMLgen.Href("#" + fullName, extra))
        return HTMLgen.Heading(2, cont, align='center')

    def generateExtraVersionHeader(self, extraVersion):
        bgColour = self.colourFinder.find("column_header_bg")
        extraVersionElement = HTMLgen.Container(HTMLgen.Name(extraVersion), extraVersion)
        columnHeader = HTMLgen.TH(extraVersionElement, colspan=len(self.tags) + 1, bgcolor=bgColour)
        return HTMLgen.TR(columnHeader)

    def escapeForHtml(self, text):
        text = html.escape(text, True)
        localeEncoding = locale.getdefaultlocale()[1] or "utf-8"
        return str(text.encode("ascii", "xmlcharrefreplace"), localeEncoding)

    def generateTestRows(self, testName, extraVersion, results):
        bgColour = self.colourFinder.find("row_header_bg")
        testId = makeTestId(self.version, extraVersion, testName)
        description = self.descriptionInfo.get(testName, "")
        container = HTMLgen.Container(HTMLgen.Name(testId), testName)
        rows = []
        testRow = [HTMLgen.TD(container, bgcolor=bgColour, title=self.escapeForHtml(description))]

        # Don't add empty rows to the table
        foundData = False
        bgcol = None
        for tag in self.tags:
            cellContent, bgcol, hasData = self.generateTestCell(tag, testName, testId, results)
            testRow.append(HTMLgen.TD(cellContent, bgcolor=bgcol))
            foundData |= hasData

        if foundData:
            # We only filter based on the final column
            self.usedColours.add(bgcol)
            rows.append(HTMLgen.TR(*testRow))
        else:
            return rows

        for resourceName in self.resourceNames:
            foundData = False
            resourceRow = [HTMLgen.TD(HTMLgen.Emphasis("(" + resourceName + ")"), align="right")]
            for tag in self.tags:
                cellContent, bgcol, hasData = self.generateTestCell(tag, testName, testId, results, resourceName)
                resourceRow.append(HTMLgen.TD(HTMLgen.Emphasis(cellContent), bgcolor=bgcol, align="right"))
                foundData |= hasData

            if foundData:
                rows.append(HTMLgen.TR(*resourceRow))
        return rows

    def getCellData(self, state, resourceName):
        if state:
            if resourceName:
                if hasattr(state, "findComparison"):
                    fileComp = state.findComparison(resourceName, includeSuccess=True)[0]
                    if fileComp:
                        return self.getCellDataFromFileComp(fileComp)
            else:
                return self.getCellDataFromState(state)

        return "N/A", True, self.colourFinder.find("test_default_fg"), self.colourFinder.find("no_results_bg")

    def getCellDataFromState(self, state):
        fileComp = state.getMostSevereFileComparison() if hasattr(state, "getMostSevereFileComparison") else None
        category = state.category if hasattr(state, "category") else "success"
        fgcol, bgcol = self.getColours(category, fileComp)
        if not hasattr(state, "category"):
            brief, hosts = parseState(state)
            return "ok " + brief + compactHostRepr(hosts), True, fgcol, bgcol
        success = category == "success"
        if success:
            cellContent = "ok"
            if state.briefText:
                cellContent += " " + state.briefText
        else:
            cellContent = state.getTypeBreakdown()[1]
        cellContent += " " + compactHostRepr(state.executionHosts)
        return cellContent.strip(), success, fgcol, bgcol

    def getCellDataFromFileComp(self, fileComp):
        success = fileComp.hasSucceeded()
        if success:
            category = "success"
        else:
            category = fileComp.getType()
        fgcol, bgcol = self.getColours(category, fileComp)
        text = fileComp.getSummary()
        return text, success, fgcol, bgcol

    def generateTestCell(self, tag, testName, testId, results, resourceName=""):
        state = results.get(tag)
        cellText, success, fgcol, bgcol = self.getCellData(state, resourceName)
        cellContent = HTMLgen.Font(cellText, color=fgcol)
        if success:
            return cellContent, bgcol, cellText != "N/A"
        else:
            linkTarget = getDetailPageName(self.pageVersion, tag) + "#" + testId
            tooltip = "'" + testName + "' failure for " + getDisplayText(tag)
            return HTMLgen.Href(linkTarget, cellContent, title=tooltip, style="color:black"), bgcol, True

    def getBackgroundColourKey(self, category):
        if category == "success":
            return "success"
        elif category == "bug":
            return "knownbug"
        elif category.startswith("faster"):
            return self.getExistingColourKey(["faster", "performance"])
        elif category.startswith("slower"):
            return self.getExistingColourKey(["slower", "performance"])
        elif category.startswith("smaller"):
            return self.getExistingColourKey(["smaller", "memory"])
        elif category.startswith("larger"):
            return self.getExistingColourKey(["larger", "memory"])
        elif category in ["killed", "unrunnable", "cancelled", "abandoned"]:
            return "incomplete"
        else:
            return "failure"

    def getExistingColourKey(self, keys):
        for key in keys:
            if self.colourFinder.find(key + "_bg"):
                return key

    def getForegroundColourKey(self, bgcolKey, fileComp):
        if (bgcolKey == "performance" and self.getPercent(fileComp) >=
            self.getConfigValue("performance_variation_serious_%", "cputime")) or \
            (bgcolKey == "memory" and self.getPercent(fileComp) >=
             self.getConfigValue("performance_variation_serious_%", "memory")):
            return "performance"
        else:
            return "test_default"

    def getColours(self, category, fileComp):
        bgcolKey = self.getBackgroundColourKey(category)
        fgcolKey = self.getForegroundColourKey(bgcolKey, fileComp)
        return self.colourFinder.find(fgcolKey + "_fg"), self.colourFinder.find(bgcolKey + "_bg")

    def getPercent(self, fileComp):
        return fileComp.perfComparison.percentageChange

    def findTagColour(self, tag):
        return self.colourFinder.find("run_" + getWeekDay(tag) + "_fg")

    def getRunNameDirs(self, repositoryDirs):
        runNameDirs = set()
        for _, dir in repositoryDirs:
            runNameDirs.add(os.path.join(os.path.dirname(os.path.dirname(dir)), "run_names"))
        return runNameDirs

    def getRunEnv(self, runEnv, key):
        return runEnv.get(key, os.getenv(key))

    def generateTableHead(self, repositoryDirs):
        head = [HTMLgen.TH("Test")]
        jenkinsUrl = os.getenv("JENKINS_URL")
        runNameDirs = self.getRunNameDirs(repositoryDirs) if jenkinsUrl else []
        for tag in self.tags:
            tagColour = self.findTagColour(tag)
            linkTarget = getDetailPageName(self.pageVersion, tag)
            linkText = HTMLgen.Font(getDisplayText(tag), color=tagColour)
            buildNumber = self.getJenkinsBuildNumber(tag)
            if jenkinsUrl and buildNumber.isdigit():
                runEnv = getEnvironmentFromRunFiles(runNameDirs, tag)
                container = HTMLgen.Container()
                tooltip = jenkinschanges.getTimestamp(buildNumber)
                container.append(HTMLgen.Href(linkTarget, linkText, title=tooltip))
                container.append(HTMLgen.BR())
                jobTarget = os.path.join(self.getRunEnv(runEnv, "JENKINS_URL"), "job",
                                         self.getRunEnv(runEnv, "JOB_NAME"), buildNumber)
                jobText = HTMLgen.Emphasis(HTMLgen.Font("(Jenkins " + buildNumber + ")", size=1))
                container.append(HTMLgen.Href(jobTarget, jobText, title=tooltip))
                head.append(HTMLgen.TH(container))
            else:
                head.append(HTMLgen.TH(HTMLgen.Href(linkTarget, linkText)))
        heading = HTMLgen.TR()
        heading = heading + head
        return heading


class TestDetails:
    def __init__(self, tag, pageTitle, pageSubTitles):
        tagText = getDisplayText(tag)
        pageDetailTitle = "Detailed test results for " + pageTitle + ": " + tagText
        self.document = HTMLgen.SimpleDocument(title=TitleWithDateStamp(pageDetailTitle), meta='<meta charset="' + locale.getpreferredencoding() + '">')
        headerText = tagText + " - detailed test results for " + pageTitle
        self.document.append(HTMLgen.Heading(1, headerText, align='center'))
        for subTitle, command in pageSubTitles:
            self.document.append(HTMLgen.Center(HTMLgen.Emphasis(subTitle)))
            self.document.append(HTMLgen.Center(HTMLgen.Paragraph(command, style='font-family:monospace')))
        self.totalCategoryHandler = CategoryHandler()
        self.versionSections = []

    def addVersionSection(self, version, categoryHandler, linkFromDetailsToOverview):
        self.totalCategoryHandler.update(categoryHandler)
        container = HTMLgen.Container()
        container.append(HTMLgen.HR())
        container.append(self.getSummaryHeading(version, categoryHandler))
        for desc, testInfo in categoryHandler.getTestsWithDescriptions():
            fullDescription = self.getFullDescription(testInfo, version, linkFromDetailsToOverview)
            if fullDescription:
                container.append(HTMLgen.Name(version + desc))
                container.append(HTMLgen.Heading(3, "Detailed information for the tests that " + desc + ":"))
                container.append(fullDescription)
        self.versionSections.append(container)

    def getSummaryHeading(self, version, categoryHandler):
        return HTMLgen.Heading(2, version + ": " + categoryHandler.generateTextSummary())

    def write(self, fileName):
        if len(self.versionSections) > 1:
            self.document.append(self.getSummaryHeading("Total", self.totalCategoryHandler))
        for sect in self.versionSections:
            self.document.append(sect)
        self.versionSections = []  # In case we get called again
        self.document.write(fileName)

    def getFreeTextData(self, tests):
        data = OrderedDict()
        for testName, state, extraVersion in tests:
            freeText = state.freeText if hasattr(state, "freeText") else None
            if freeText:
                if freeText not in data:
                    data[freeText] = []
                data[freeText].append((testName, state, extraVersion))
        return list(data.items())

    def getFullDescription(self, tests, version, linkFromDetailsToOverview):
        freeTextData = self.getFreeTextData(tests)
        if len(freeTextData) == 0:
            return
        fullText = HTMLgen.Container()
        for freeText, tests in freeTextData:
            tests.sort(key=lambda info: info[0])
            for testName, _, extraVersion in tests:
                fullText.append(HTMLgen.Name(makeTestId(version, extraVersion, testName)))
            fullText.append(self.getHeaderLine(tests, version, linkFromDetailsToOverview))
            self.appendFreeText(fullText, freeText)
            if len(tests) > 1:
                for line in self.getTestLines(tests, version, linkFromDetailsToOverview):
                    fullText.append(line)
        return fullText

    def appendFreeText(self, fullText, freeText):
        freeText = freeText.replace("<", "&lt;").replace(">", "&gt;")
        linkMarker = "URL=http"
        if linkMarker in freeText:
            currFreeText = ""
            for line in freeText.splitlines():
                if linkMarker in line:
                    fullText.append(HTMLgen.RawText("<PRE>" + currFreeText.strip() + "</PRE>"))
                    currFreeText = ""
                    words = line.strip().split()
                    linkTarget = words[-1][4:]  # strip off the URL=
                    newLine = " ".join(words[:-1])
                    fullText.append(HTMLgen.Href(linkTarget, newLine))
                    fullText.append(HTMLgen.BR())
                else:
                    currFreeText += line + "\n"
        else:
            currFreeText = freeText
        if currFreeText:
            fullText.append(HTMLgen.RawText("<PRE>" + currFreeText + "</PRE>"))

    def getHeaderLine(self, tests, version, linkFromDetailsToOverview):
        testName, state, extraVersion = tests[0]
        if len(tests) == 1:
            linksToOverview = self.getLinksToOverview(version, testName, extraVersion, linkFromDetailsToOverview)
            headerText = "TEST " + repr(state) + " " + testName + " ("
            container = HTMLgen.Container(headerText, linksToOverview)
            return HTMLgen.Heading(4, container, ")")
        else:
            headerText = str(len(tests)) + " TESTS " + repr(state)
            return HTMLgen.Heading(4, headerText)

    def getTestLines(self, tests, version, linkFromDetailsToOverview):
        lines = []
        for testName, _, extraVersion in tests:
            linksToOverview = self.getLinksToOverview(version, testName, extraVersion, linkFromDetailsToOverview)
            headerText = testName + " ("
            container = HTMLgen.Container(headerText, linksToOverview, ")<br>")
            lines.append(container)
        return lines

    def getLinksToOverview(self, version, testName, extraVersion, linkFromDetailsToOverview):
        links = HTMLgen.Container()
        for targetFile, linkName in linkFromDetailsToOverview:
            testId = makeTestId(version, extraVersion, testName)
            links.append(HTMLgen.Href(targetFile + "#" + testId, linkName))
        return links


class CategoryHandler:
    def __init__(self):
        self.testsInCategory = OrderedDict()

    def update(self, categoryHandler):
        for category, testInfo in list(categoryHandler.testsInCategory.items()):
            testInfoList = self.testsInCategory.setdefault(category, [])
            testInfoList += testInfo

    def registerInCategory(self, testId, category, extraVersion, state):
        self.testsInCategory.setdefault(category, []).append((testId, state, extraVersion))

    def getDescription(self, cat, count):
        shortDescr, _ = getCategoryDescription(cat)
        return str(count) + " " + shortDescr

    def getTestCountDescription(self, count):
        return str(count) + " tests: "

    def generateTextSummary(self):
        numTests, summaryData = self.getSummaryData()
        categoryDescs = [self.getDescription(cat, count) for cat, count in summaryData]
        return self.getTestCountDescription(numTests) + " ".join(categoryDescs)

    def generateHTMLSummary(self, detailPageRef, extraVersion=None):
        numTests, summaryData = self.getSummaryData(extraVersion)
        container = HTMLgen.Container()
        for cat, count in summaryData:
            summary = HTMLgen.Text(self.getDescription(cat, count))
            if cat == "success":
                container.append(summary)
            else:
                linkTarget = detailPageRef + getCategoryDescription(cat)[-1]
                container.append(HTMLgen.Href(linkTarget, summary))

        testCountSummary = HTMLgen.Text(self.getTestCountDescription(numTests))
        return HTMLgen.Container(testCountSummary, container)

    def countTests(self, testInfo, extraVersion):
        if extraVersion is not None:
            return sum((currExtra == extraVersion for (testId, state, currExtra) in testInfo))
        else:
            return len(testInfo)

    def getSummaryData(self, extraVersion=None):
        numTests = 0
        summaryData = []
        for cat, testInfo in list(self.testsInCategory.items()):
            testCount = self.countTests(testInfo, extraVersion)
            if testCount > 0:
                summaryData.append((cat, testCount))
                numTests += testCount
        summaryData.sort(key=self.getSummarySortKey)
        return numTests, summaryData

    def getTestsWithDescriptions(self):
        return sorted([(getCategoryDescription(cat)[1], testInfo) for cat, testInfo in list(self.testsInCategory.items())])

    def getSummarySortKey(self, data):
        # Put success at the start, it's neater like that
        return data[0] != "success", -data[1], data[0]


def getCategoryDescription(cat):
    return plugins.TestState.categoryDescriptions.get(cat, (cat, cat))


def getDetailPageName(pageVersion, tag):
    return "test_" + pageVersion + "_" + tag + ".html"


class BaseSelector(object):
    def __init__(self, linkName, suffix):
        self.selectedTags = []
        self.linkName = linkName
        self.suffix = suffix

    def add(self, tag):
        self.selectedTags.append(tag)

    def getLinkInfo(self, pageVersion):
        return "test_" + pageVersion + self.suffix + ".html", self.linkName

    def sortKey(self):
        pass


class Selector(BaseSelector):
    def __init__(self, linkName, suffix, getConfigValue, tags):
        super(Selector, self).__init__(linkName, suffix)
        cutoff = getConfigValue("historical_report_subpage_cutoff", linkName)
        weekdays = getConfigValue("historical_report_subpage_weekdays", linkName)
        self.selectedTags = tags
        if len(weekdays) > 0:
            self.selectedTags = [tag for tag in self.selectedTags if getWeekDay(tag) in weekdays]
        self.selectedTags = self.selectedTags[-cutoff:]


class SelectorByMonth(BaseSelector):
    @classmethod
    def makeInstances(cls, tags):
        allSelectors = {}
        for tag in tags:
            month = tag[2:9]
            allSelectors.setdefault(month, SelectorByMonth(month)).add(tag)
        return sorted(allSelectors.values(), key=lambda s: s.sortKey())

    def __init__(self, month):
        super(SelectorByMonth, self).__init__(month, "_all_" + month)
        self.month = datetime.strptime(self.linkName, "%b%Y")

    def getPreviousMonthSelector(self):
        first_day_of_current_month = self.month.replace(day=1)
        last_day_of_previous_month = first_day_of_current_month - timedelta(days=1)
        return SelectorByMonth(last_day_of_previous_month.strftime("%b%Y"))

    def sortKey(self):
        return self.month

    def __eq__(self, other):
        return self.linkName == other.linkName

    def __hash__(self):
        return self.linkName.__hash__()




import shutil
import datetime
import time
import os
from texttestlib import plugins
from . import vcs_independent


class HgInterface(vcs_independent.VersionControlInterface):
    def __init__(self, controlDir):
        self.warningStateInfo = {"M": "Modified", "R": "Removed", "A": "Added"}
        self.errorStateInfo = {"?": "Unknown", "!": "Missing"}
        self.allStateInfo = {"C": "Unchanged", "I": "Ignored"}
        self.allStateInfo.update(self.warningStateInfo)
        self.allStateInfo.update(self.errorStateInfo)
        vcs_independent.VersionControlInterface.__init__(self, controlDir, "Mercurial",
                                                         list(self.warningStateInfo.values()), list(self.errorStateInfo.values()), "tip")
        self.defaultArgs["rm"] = ["--force"]
        self.defaultArgs["status"] = ["-A"]  # Otherwise we can't tell difference between Unchanged and Ignored
        self.defaultArgs["log"] = ["-f"]  # "follow" renames, which isn't default
        self.defaultArgs["annotate"] = ["-f", "-n"]  # annotate -f doesn't annotate anything...
        self.defaultArgs["diff"] = ["-g"]  # "git format", which apparently is how to make it work with revisions :)
        self.defaultArgs["pull"] = ["-u"]  # Otherwise doesn't actually update the files. We "pretend to be centralised"

    def getDateFromLog(self, output):
        for line in output.splitlines():
            if line.startswith("date:"):
                dateStr = " ".join(line.split()[2:-1])
                return datetime.datetime(*(self.parseDateTime(dateStr)[0:6]))

    def getGraphicalDiffArgs(self, diffProgram):
        return ["hg", "extdiff", "-p", diffProgram]

    def parseDateTime(self, input):
        return time.strptime(input, "%b %d %H:%M:%S %Y")

    def getFileStatus(self, basicArgs, file):
        return vcs_independent.VersionControlInterface.getFileStatus(self, basicArgs, self.correctForLinks(file))

    def correctForLinks(self, file):
        # Mercurial falls over if "file" contains symbolic link references
        # But don't dereference it if it's actually a link
        if os.path.islink(file):
            return file
        else:
            return os.path.realpath(file)

    def callProgramOnFiles(self, cmdName, fileArg, recursive=False, extraArgs=[], **kwargs):
        basicArgs = self.getCmdArgs(cmdName, extraArgs)
        for fileName in self.getFileNamesForCmd(cmdName, fileArg, recursive):
            self.callProgramWithHandler(fileName, basicArgs + [self.correctForLinks(fileName)], **kwargs)

    def getStateFromStatus(self, output):
        words = output.split()
        if len(words) > 0:
            statusLetter = words[0]
            return self.allStateInfo.get(statusLetter, statusLetter)
        else:  # pragma: no cover - robustness fix only. Could be triggered before 'correctForLinks' was introduced
            # We don't really know, but we assume an error message means we probably don't control the file
            return "Unknown"

    def _movePath(self, oldPath, newPath):
        # Moving doesn't work in hg if there are symbolic links in the path to the new location!
        retCode = self.callProgram("mv", [os.path.realpath(oldPath), os.path.realpath(newPath)])
        # And it doesn't take non-versioned files with it, if there are any...
        if retCode == 0 and os.path.isdir(oldPath):
            self.copyPath(oldPath, newPath)
            shutil.rmtree(oldPath)

    def removePath(self, path):
        # Mercurial doesn't remove unknown files, nor does it return non-zero exit codes when it fails
        retValue = os.path.exists(path)
        self.callProgram("rm", [self.correctForLinks(path)])  # returns 0 whatever happens
        plugins.removePath(path)
        return retValue

    def hasLocalCommits(self, vcsDirectory):
        return self.callProgram("out", cwd=vcsDirectory) == 0


class UpdateGUI(vcs_independent.UpdateGUI):
    def getCommandName(self):
        return "pull"

    @staticmethod
    def _getTitle():
        return "Pull/Update"


vcs_independent.vcsClass = HgInterface


class AnnotateGUI(vcs_independent.AnnotateGUI):
    def commandHadError(self, retcode, stderr, *args):
        return len(stderr) > 0  # Mercurial doesn't do return codes in annotate for some reason...


class AnnotateGUIRecursive(AnnotateGUI):
    recursive = True

#
# Configuration for the Interactive Actions
#


class InteractiveActionConfig(vcs_independent.InteractiveActionConfig):
    def annotateClasses(self):
        return [AnnotateGUI, AnnotateGUIRecursive]

    def getUpdateClass(self):
        return UpdateGUI
